#include "isubsys.h"
#include "pbotserialparser.h"
#include "simpletext.h"
#include "simpletools.h"

PBotSerialParser::PBotSerialParser()
{
    ready = 0;
    termCnt = 0;
    clrTerms();
    clrSubSys();
}


PBotSerialParser::PBotSerialParser(int rxPin, int txPin, int b_rate)
{
    ready = 0;
    termCnt = 0;
    clrTerms();
    clrSubSys();
    init(rxPin, txPin, b_rate);
}


int PBotSerialParser::init(int rxPin, int txPin, int b_rate)
{
    low(15);
    p_cmdTerm = fdserial_open(rxPin,txPin,0,b_rate);

    if (p_cmdTerm)
    {
        ready = 1;
    }

    return ready;
}


/** @brief Indicates if the object is ready to parse.
  *
  * @return int: Boolean ready status.
  */
int PBotSerialParser::isReady()
{
    return ready;
}


/** @brief (one liner)
  *
  * @return int: Number of characters in the command. Negative value indicates error.
  */
int PBotSerialParser::checkCmd()
{
    char command[MAX_CMD_SIZE+1];   // Command being assembled
    int index = 0;                  // Character position being evaluated
    int startCnt = 0;               // Mark time for waiting for complete command
    int result = -1;                
 
    if (ready < 1)
        return CMD_PARSER_ERR_GEN;
    
    result = fdserial_rxCheck(p_cmdTerm);

    if (result == CMD_START)
    {
        // We have a valid command start.  Begin loop to process incomming characters.
        clrTerms();                 // Clear any existing terms in the object array
        terms[0] = -1;              // Start position is one less than position 0
        termCnt = 1;                // Start with one term
        
        startCnt = CNT;             // Get the time when we started waiting for the next character
        while(1)
        {
            result = -1;                // Reset the result value
        
            while (result < 0)          // Wait for the next character
            {
                result = fdserial_rxCheck(p_cmdTerm);
            }

            if (index > MAX_CMD_SIZE)       // If the command is too long then abort and indicate error.
            {
                writeChar(p_cmdTerm, ERR_CHAR);
                return CMD_PARSER_ERR_GEN;
            }
            
            if (result == TERM_DELIM)       
            {
                terms[termCnt] = index;     // Mark term delim positions if found
                termCnt++;
            }

            if (result == CMD_END)          // if the termination character is detected then finish the string,
            {
                int validate = 0;

                command[index] = '\0';
                rawCommand = command;

                validate = dispCmd();       // Dispatch the command to the requested subsystem
                if(validate < 0)
                {
                    writeChar(p_cmdTerm, ERR_CHAR);
                    return CMD_PARSER_ERR_GEN;
                }
                
                writeChar(p_cmdTerm, ACK_CHAR);

                if(validate > 0)
                    sendResp();

                return index;
            }

            command[index] = result;
            index++;
        }
    }

    // No command start character detected.
    return CMD_PARSER_ERR_GEN;
}


/** @brief Dispatch the command to the proper subsystem
  *
  * @return int: Negative value indicates error.  All other return values indicate success.
  */
int PBotSerialParser::dispCmd()
{
    // Iterate through the subsystem name list and look for one matching the first term.
    // Message that subsystem and inform it that a command is ready for it.  Acknowlege
    // successful dispatch to the host.
    int subsysIndex = -1;
    int validate = -1;
    String nameTerm;

    // If no initial term is found return error
    if (getTerm(nameTerm, 0) < 0)
        return CMD_PARSER_ERR_GEN;

    
    // Seek to the name of the subsystem
    for(int i=0; i< MAX_SUBSYS; i++)
    {
        if(subSysName[i] == nameTerm)
        {
            subsysIndex = i;
            break;
        }
    }

    // If no matching subsystem return error
    if (subsysIndex < 0)
        return CMD_PARSER_ERR_GEN;

    // Dispatch to the matching subsystem
    validate = p_subSys[subsysIndex]->execCmd();
    return validate;
}


/** @brief Clears all term markers in the term database.
  */
void PBotSerialParser::clrTerms()
{
    termCnt = 0;
    for(int i=0; i<MAX_TERMS; i++)
    {
        terms[i] = -255;
    }
}


/** @brief Clears all subsystem pointers from the array.
  */
void PBotSerialParser::clrSubSys()
{
    for(int i=0; i<8; i++)
    {
        p_subSys[i] = 0;
    }
}


/** @brief Return the entire command string
  *
  * (documentation goes here)
  */
char* PBotSerialParser::getCmd()
{
    return rawCommand.c_str();
}


/** @brief Locate the requested term from the comamnd string and place it in provided reference
  *
  * @param String& term: A String object in which to store the fetched term.
  * @param int termIndex: The index of the term you wish to retreive.
  * @return int: zero for success, negative value on failure
  */
int PBotSerialParser::getTerm(String &term, int termIndex)
{
    int startPos = 0;
    int endPos = 0;

    if (termIndex > MAX_TERMS)
    {
        term = "ERR";
        return CMD_PARSER_ERR_GEN;
    }
    
    startPos = terms[termIndex];    // Start position is at the requested term index
    endPos = terms[termIndex+1];    // End position is at the next term index
    
    if (startPos == -255)
    {
        term = "ERR";
        return CMD_PARSER_ERR_GEN;
    }

    if (endPos == -255)
    {
        endPos = rawCommand.length();
    }

    term = rawCommand.substring(startPos+1, endPos);
    return 0;
}


/** @brief Return the count of the number of terms for the current command
  *
  * @return int: The number of terms in the last processed command.
  */
int PBotSerialParser::getTermCnt()
{
    return termCnt-1;
}


/** @brief sends a response on the TX line to the controlling host
  *
  * @return int: Zero on sucess, negative number on failure.
  */
int PBotSerialParser::sendResp()
{
    int subsysIndex = -1;
    String response = "4321";
    String strOut = CMD_START;
    String nameTerm = "";

    // If no initial term is found return error
    if (getTerm(nameTerm, 0) < 0)
        return CMD_PARSER_ERR_GEN;

    // Seek to the name of the subsystem
    for(int i=0; i< MAX_SUBSYS; i++)
    {
        if(subSysName[i] == nameTerm)
        {
            subsysIndex = i;
            break;
        }
    }

    // If no matching subsystem return error
    if (subsysIndex < 0)
        return CMD_PARSER_ERR_GEN;

    // Dispatch to the matching subsystem
    response = p_subSys[subsysIndex]->getResult();
    strOut += response;
    strOut += CMD_END;

    writeStr(p_cmdTerm, strOut.c_str());
    putStr(strOut.c_str());
    putStr("\n");
    return 0;
}


/** @brief Add a subsytem to the list for command dispatching
  *
  * Stores a new SubSytem object in the array and associates it with a string identifier.  This object
  * does not take responcibilty for storage of the actual object.  The object must remain valid and
  * in-scope for as long as it is added.  SubSystems are responcible for handeling commands that are
  * addressed to their string identifier.  The first term of any command is the SubSystem string ID
  * that will handle that command.
  *
  * @return int: Index the SubSystem was stored at.  Negative number for error.
  */
int PBotSerialParser::addSubSys(ISubSys& subSys, const char* name)
{
    // Locate the first index availible to store the Subsystem pointer and place it there
    int i = 0;
    while (p_subSys[i] != 0)
    {
        i++;
        if (i > MAX_SUBSYS)
        {
            return CMD_PARSER_ERR_GEN;
        }
    }

    p_subSys[i] = &subSys;
    subSysName[i] = name;
    p_subSys[i]->attachParser(this);
    return i;
}



